#include <EEPROM.h>
// #include <SoftwareSerial.h>
#include <OneWire.h>
#include <PID_v1.h>       
#include <DS1307.h>  

#define LEDOUTPUT 6		// Светодиод повторяет мощность выхода
#define LED_WORK 13		// Режим работы
#define TIME_TO_UPDATE_NEXTION 5000 // Обновляем информацию на дисплее
#define PIN_HUMI A0

//---------------  Attention!!!--------
#define DAY_LONG_SECONDS 86400
//#define DAY_LONG_SECONDS 15
unsigned  long dayLastsSeconds = DAY_LONG_SECONDS;
//-------------------------------------

OneWire  ds(2);
byte addr[2][8];

// SoftwareSerial NextSerial(10, 11); // RX, TX
String inStr = ""; 
uint8_t inByte;
bool serialReadFlag = false;

uint8_t unitWorksFlag = 0;		// Режим: 0 - стоп, 1 - работаем, 2 - auto

double setpoint, input, output;
double aggKp=4, aggKi=0.2, aggKd=1;       // Определяем агрессивные и консервативные параметры PID;
double consKp=1, consKi=0.05, consKd=0.25;
double midKp=1.5, midKi=0.1, midKd=0.65;

// Определяем указатели и другие параметры PID;
PID myPID(&input, &output, &setpoint, consKp, consKi, consKd, DIRECT);
double gap;                     		// Переключатель параметров PID;

//=========================
volatile unsigned int myPWM = 5;  	// Выключено, переменная для управления мощностью 
volatile uint8_t sreg;				// Переменная для запоминания регистра SREG

DS1307 rtc(SDA, SCL);
Time  t;							// Переменная для чтения в нее времени;
uint32_t unixStart;					// Начало отсчета рабочих режимов

uint8_t dayOfDuty;					// Текущий день работы
uint8_t autoM[20];					// Массив смены температур для автомата

unsigned  long timeToSend = 0;

const PROGMEM String stopS   = "\xc1\xe2\xde\xdf"; // Стоп
const PROGMEM String manualS = "\xc0\xe3\xe7\xdd\xde\xd9";  // Ручной
const PROGMEM String autoS   = "\xb0\xd2\xe2\xde";  // Авто
const PROGMEM String varn    = "Tmp problem!";
bool  recountNow = true;
bool  firstDS18b20 = false;
bool  secondDS18b20 = false;

float humi;

//========================  Functions
ISR(TIMER1_COMPA_vect ){			// Отработка прерывания Timer1 по сравнению с регистром OC1A
	PORTC |= ( 1 << PC1 );			// Установка A1 в 1
}
ISR(INT1_vect) {
	PORTC &= ~( 1 << PC1 );			// Сброс A1 в 0
	sreg = SREG;					// Запомнили  состояние общего флага прерываний
	cli();
	TCNT1 = myPWM;
	SREG = sreg;
}
void	UpdatePID(double targetTemperature) {
		
	gap = abs(setpoint-targetTemperature);      // Разность с желаемой температурой
	
	/*if(gap<5) {                 				// Мы недалеко и используем консервативные параметры;
		myPID.SetTunings(consKp, consKi, consKd);
	}
	else {               						 // Мы далеко, используем агрессивные параметры;
		myPID.SetTunings(aggKp, aggKi, aggKd);
	}
	*/
	myPID.Compute();
		
	if( (setpoint - input) > 5) {
		myPWM = 159;
	}
	else if( (input - setpoint) > 5) {
		myPWM = 6;
	}
	else {
		myPWM = (uint8_t)output;
	}
	analogWrite(LEDOUTPUT, (byte)(myPWM*1.58)); 
}
void printNextion(const char*, String, String);

void GetRH(float *cell) {
  float RH = (((5.0/1024*analogRead(PIN_HUMI))/5.0) - 0.16)/0.0062; 
  if(RH > 99) RH = 99;
  *cell = RH;
}

void setup() {
  	SetTimer();
	rtc.begin();
	rtc.halt(false);              	// Запуск часов;
	
	firstDS18b20  = ds.search(addr[0]);             // Найдем и запомним адреса  двух DS18B20;
	secondDS18b20 = ds.search(addr[1]);
	if(!(firstDS18b20 && secondDS18b20)) {
		printNextion(varn.c_str(), "t7", "0");
	}
	
	Serial.begin(115200);
	
	myPID.SetMode(AUTOMATIC);       // Включим PID и установим режим;
	myPID.SetOutputLimits(10, 158);
	// myPID.SetTunings(aggKp, aggKi, aggKd);
	//myPID.SetTunings(consKp, consKi, consKd);
	myPID.SetTunings(midKp, midKi, midKd);
	pinMode(LED_WORK, OUTPUT);		// Индикация режима работы;
	
	unitWorksFlag	= EEPROM.read(0);  // Прочитали режим

	setpoint = (double) EEPROM.read(1); // Прочитали записанную целевую температуру
	if (setpoint>120.0 || setpoint < 70.0) setpoint = 80.0;	
	
	EEPROM.get(2, unixStart);  // Прочитали время старта активных режимов  
	
	for(uint8_t i = 0; i < 20; i++) { // Прочитали график смены температуры
		autoM[i] = EEPROM.read(i+10);
	}
	
	CountDays();  // Подсчитали  день от начала активного режима
	NextUpdateMode(unitWorksFlag); // Вывели результат на дисплей
	
	// Установка часов на всякий случай
	/*
	rtc.setDOW(SUNDAY);        // Set Day-of-Week to SUNDAY
	rtc.setTime(19, 20, 0);     // Set the time to 12:00:00 (24hr format)
	rtc.setDate(11, 13, 2016);   // Set the date to mohth, day, year.
	*/
}

void loop() {
	input = (double) CheckTemperature();
	if(unitWorksFlag) {         			// Если работаем - обновляем выход PID
		UpdatePID(input);					
		// digitalWrite(LED_WORK,1);			// и жгем лампу
	}
	/* else {									// Иначе - все выключаем 
		myPWM = 10;
		// digitalWrite(LED_WORK,0);
	}
	*/
	if (Serial.available()) {
		uint8_t inn = Serial.read();
		if(serialReadFlag) {
			if(inn == 59){     // ASCII : ";"
				checkCommand(inStr);
				serialReadFlag = false;
			}
			else {
				inStr += (char)inn;
				inStr.trim();
			}
		}
		else {
			if(inn == 35) { // ASCII : "#"
				serialReadFlag = true;
				inStr = "";
			}
		}
	}

	if(millis() > timeToSend) {
		timeToSend = millis() + TIME_TO_UPDATE_NEXTION;
		CountDays();
		NextUpdateMode(unitWorksFlag);
		// Serial.println(rtc.getTimeStr());
	}
}

void checkCommand(String ins) {
	String first = ins.substring(0,1);
	String last  = ins.substring(1);
  
	if(first == "t") { // 't'  ASCII Value: 116
		float ttt = last.toFloat();
		if((ttt > 36.5) && (ttt < 42.5)) {
			setpoint = (double) ((ttt*10) - 300); // работает с температурой типа ((38.5 * 10) - 300) = 85
			EEPROM.update(1,(byte) setpoint );
		} 
	}

	if(first == "d") { // 'd'  ASCII Value: 100
		
		uint8_t dd = last.toInt();

		((dd > 0) && (dd< 30)) ? dayOfDuty = dd : 1 == 1;
		printNextion(last.c_str(), "t4", "0");
		//Serial.print("Day - 1 = ");
		//Serial.println(dd - 1);
		SetStartUnixTime();
	}
	   
	if(first == "m") { // 'm'  ASCII Value: 109
		uint8_t mm = (uint8_t)last.toInt();
		if((mm >= 0) && (mm <3)) {
			unitWorksFlag = mm;
			EEPROM.update(0, unitWorksFlag);
			if (unitWorksFlag == 0) myPWM = 5;
			printNextion(((String)((setpoint + 300)/10.0)).c_str(), "t2", "0"); // Температура	
		}
	}

	if(first == "a") { 	// 'a'  ASCII Value: 97
		fillAutoM();  	// Обнулим все элементы массива
		uint8_t length = last.length();
		uint8_t place = 0;
		String dat = "";
		//Serial.println("Got new auto data");
			
		for(uint8_t i = 0; i < length; i++) {
			String let = last.substring(i, i+1);
			if(let !="," && let !=":") {
				dat = dat + let;
				continue;
			}
			float dig = dat.toFloat();
			if(dig < 30) {
				autoM[place] = (uint8_t)dig;
				//Serial.print(autoM[place]);
				//Serial.print(":");
			} 
			else {
				dig = dig * 10;
				int tt = ((int)dig) - 300;
				autoM[place] = (uint8_t)tt;
				
				//Serial.print(autoM[place]);
				//Serial.print(",");
			}
			place++;
			dat = "";
		}

		for(uint8_t i = 0; i < 20; i++) {
			EEPROM.update(i+10, autoM[i]);
			Serial.print(autoM[i]);
			Serial.print(" ");
		}
	}

	if(first == "g") { // 'g'  ASCII Value: 103
		// bool flag = true;
		String forNext = "";
		for(uint8_t i =0; i <20; i=i+2) {
			uint8_t d = autoM[i];
			if(d !=0) {
				forNext = forNext + String(d) + ":";
				int tm = ((int)autoM[i+1]) + 300;
				float fm = ((float)tm) / 10;
				forNext = forNext + String(fm) + ",";
			}
		}
		
		printNextion(forNext.c_str(),"t2","4");
		Serial.print("Auto = ");
		Serial.println(forNext);
		
	}
	
	if(first == "u"){
		dayLastsSeconds = (unsigned long) last.toInt();
		printNextion(((String)dayLastsSeconds).c_str(), "t1", "6");
	}
	if(first == "x"){
		printNextion(((String)dayLastsSeconds).c_str(), "t1", "6");
	}
	
	
	NextUpdateMode(unitWorksFlag); 
}

	// ----- Установка Unix времени начала работы
void SetStartUnixTime(void){
	t=rtc.getTime();
	unixStart = rtc.getUnixTime(t) - (uint32_t)(DAY_LONG_SECONDS * (dayOfDuty - 1));
	EEPROM.put(2, unixStart);
	recountNow = true;
	CountDays(); 
}


void printNextion(const char* data, String top, String page){
	String spage = "page"+page+"."+top+".txt=\"";
	Serial.print(spage);
	Serial.print(data);
	Serial.print("\"");
	Serial.write(0xFF);
	Serial.write(0xFF);
	Serial.write(0xFF);
}

void commandNextion(const char* data){
	Serial.print(data);
	Serial.write(0xFF);
	Serial.write(0xFF);
	Serial.write(0xFF);
}

void fillAutoM(void) {
	for (uint8_t i = 0; i < 20; i++){
		autoM[i] = 0;
		EEPROM.update(i+10,0);
	}
} 

void NextUpdateMode(uint8_t mod) {
	GetRH(&humi);
	printNextion((((String)((uint8_t)humi))+"%").c_str(), "t6", "0");
	
	static float ssetpoint = 0;
	static uint8_t ddayOfDuty = 30;
	static uint8_t mmod = 3;
		
	String modd = "";
	
	if(mod !=0) {
		float tmp = (setpoint + 300)/10;
		if(ssetpoint != setpoint){
			ssetpoint = setpoint;
			printNextion(((String)tmp).c_str(), "t2", "0"); // Температура	
		}
		if(ddayOfDuty != dayOfDuty){
			ddayOfDuty = dayOfDuty;
			printNextion(((String)(dayOfDuty)).c_str(), "t4", "0"); // День
		}
	}
	
	if(mmod != mod){
		mmod = mod;
		switch(mod){
			case 0:
				modd = stopS;
				printNextion("~", "t2", "0"); // Температура
				printNextion("x", "t4", "0"); // День
				break;
			case 1:
				modd = manualS;
				break;
			case 2:
				modd = autoS;
				break;
		}
		printNextion(((String)modd).c_str(), "t3", "0");// Режим
	}
}

void CountDays() {
	static uint8_t dayNow;  // Переменная постоянно пересчитывает время от Unix-переменной 
	static bool changeDayFlag = false;
	t=rtc.getTime();
	dayNow = ((uint8_t)((rtc.getUnixTime(t) - unixStart) / dayLastsSeconds)) + 1; // Это день от записанного Unix
	
	
	//---------------  Attention!!!
	if (dayNow >= 25) {
		dayLastsSeconds = 86400;
	}
	//---------------
	if ((((dayNow != dayOfDuty) && (unitWorksFlag))) || (recountNow == true)) {
		recountNow = false;
		dayOfDuty =  dayNow;
		changeDayFlag = false;
		// Serial.println("Recount now!");
		
		if ((unitWorksFlag == 2) && !changeDayFlag) {
			uint8_t a;
			// Serial.println("Yes, got change!");
			for(a = 0; autoM[a]; a = a+2) {
				// Serial.print("Ask Day: ");
				// Serial.println(autoM[a]);
				if(autoM[a] >= (dayOfDuty - 1)) {
					setpoint = autoM[a+1];
					// Serial.print("Got t: ");
					// Serial.println(setpoint);
					EEPROM.update(1,(byte)(setpoint - 300.0));
					// Serial.print("Update setpoint: ");
					// Serial.println(setpoint);
					break;
				}
			}
			changeDayFlag = true;
		}
	}
}


float	CheckTemperature(void) {      				// Проверка температуры  и выдача среднего значения на PID
	static unsigned long timeWhenAskTemperature;
	static uint8_t stepCheck = 0;		
	static float tempNow01;          				// Текущая Температура 1
	static float tempNow02;          				// Текущая Температура 2
	if (millis() > timeWhenAskTemperature) {
		String t1 = "";
		timeWhenAskTemperature = millis() + 200;	// Шаги по времени 
		switch (stepCheck) {
			case 0:
				AskTemperature(addr[0]);			// Запрос температуры 1
				stepCheck++;
			break;

			case 1:
				if(secondDS18b20) {
					AskTemperature(addr[1]);			// Запрос температуры 2
				}
				stepCheck++;
				
			break;

			case 2:									// Получение температуры 1
				tempNow01 = (float)GetTemperature(addr[0]);
				stepCheck++;
				t1 = (String)(tempNow01/100.0);
				printNextion(t1.c_str(), "t0", "0");
				
			break;
			 
			case 3:									// Получение температуры 2
				if(secondDS18b20) {
					tempNow02 = (float)GetTemperature(addr[1]);
				}
				else {
					tempNow02 = tempNow01;
				} 
					
				stepCheck = 0;
				t1 = (String)(tempNow02/100.0);
				printNextion(t1.c_str(), "t1", "0");
			break;
		} 
	}
	float mid = (tempNow01 + tempNow02)/200;
	String tm = (String)mid;
	
	printNextion(tm.c_str(), "t5", "0");
	
	float middle = ((tempNow01 + tempNow02)/20) - 300;		// Выдача средней температуры
	/*
	Serial.print("Middle temp = ");
	Serial.println(middle);
	Serial.print("Setpoint temp = ");
	Serial.println(setpoint);
	*/
	return middle;
	// return (mid - 300);

}
void AskTemperature(byte* addressNow) {			// Запрос температуры 
	byte * addrr = addressNow;
	ds.reset();                   					
    ds.select(addrr);
    ds.write(0x44,1);           
}
int	GetTemperature(byte* addressNow) {				// Получение значения температуры
	byte * addrr = addressNow;
	int highByte, lowByte, tReading, tc_100;
    ds.reset();
    ds.select(addrr);
    ds.write(0xBE);             					// Указание на будущее чтение;
 	lowByte  = ds.read();
	highByte = ds.read();
    tReading = (highByte << 8) + lowByte;
    return tc_100 = (6 * tReading) + tReading / 4; // Стандартная операция пересчета для DS18b20 (100 * 0.0625) или 6.25;
}

/**/
void SetTimer(void) {							// Инициализация таймера
	// Port C initialization   PC1 - A1 Arduino
	// Function: Bit6=In Bit5=In Bit4=In Bit3=In Bit2=In Bit1=Out Bit0=In
	DDRC=(0<<DDC6) | (0<<DDC5) | (0<<DDC4) | (0<<DDC3) | (0<<DDC2) | (1<<DDC1) | (0<<DDC0);
	
	// 160 = 0xA0 - 10 ms
	// Timer/Counter 1 initialization
	// Clock source: System Clock
	// Clock value: 15,625 kHz
	// Mode: Normal top=0xFFFF
	// OC1A output: Disconnected
	// OC1B output: Disconnected
	// Noise Canceler: Off
	// Input Capture on Falling Edge
	// Timer Period: 4,1943 s
	// Timer1 Overflow Interrupt: Off
	// Input Capture Interrupt: Off
	// Compare A Match Interrupt: On
	// Compare B Match Interrupt: Off
	
	TCCR1A=(0<<COM1A1) | (0<<COM1A0) | (0<<COM1B1) | (0<<COM1B0) | (0<<WGM11) | (0<<WGM10);
	TCCR1B=(0<<ICNC1) | (0<<ICES1) | (0<<WGM13) | (0<<WGM12) | (1<<CS12) | (0<<CS11) | (1<<CS10);
	TCNT1H=0x00;
	TCNT1L=0x00;
	ICR1H=0x00;
	ICR1L=0x00;
	OCR1AH=0x00;
	OCR1AL=0xA0;
	OCR1BH=0x00;
	OCR1BL=0x00;
	
	// Timer/Counter 1 Interrupt(s) initialization
	TIMSK1=(0<<ICIE1) | (0<<OCIE1B) | (1<<OCIE1A) | (0<<TOIE1);
	
	// External Interrupt(s) initialization
	// INT0: Off
	// INT1: On
	// INT1 Mode: Rising Edge
	// Interrupt on any change on pins PCINT0-7: Off
	// Interrupt on any change on pins PCINT8-14: Off
	// Interrupt on any change on pins PCINT16-23: Off
	
	EICRA=(1<<ISC11) | (1<<ISC10) | (0<<ISC01) | (0<<ISC00);
	EIMSK=(1<<INT1) | (0<<INT0);
	EIFR=(1<<INTF1) | (0<<INTF0);
	PCICR=(0<<PCIE2) | (0<<PCIE1) | (0<<PCIE0);
	
	sei();  // Разрешаем прерывания
}

/**/
